<!DOCTYPE html>
<html lang="en">
<head>
    <!-- Required meta tags -->
    <meta charset="utf-8"/>
    <meta name="warehub:repository-version" content="%%WAREHUB_VERSION%%"/>
    <meta content="width=device-width, initial-scale=1, shrink-to-fit=no" name="viewport"/>
    <!-- Favicon -->
    <link href="%%IMAGE%%" rel="icon"/>
    <title>%%TITLE%%</title>
</head>
<body>
<div class="container">%%LIST%%
</div>
</body>
</html>
