<!DOCTYPE html>
<html lang="en">
<head>
    <!-- Required meta tags -->
    <meta charset="utf-8"/>
    <meta name="warehub:repository-version" content="%%WAREHUB_VERSION%%"/>
    <meta content="width=device-width, initial-scale=1, shrink-to-fit=no" name="viewport"/>
    <!-- Skeleton CSS -->
    <link crossorigin="anonymous" href="https://cdnjs.cloudflare.com/ajax/libs/skeleton/2.0.4/skeleton.min.css" rel="stylesheet"/>
    <!-- Font -->
    <link href="https://fonts.googleapis.com/css2?family=Montserrat:wght@300;400;600&amp;display=swap" rel="stylesheet" type="text/css"/>
    <!-- Favicon -->
    <link href="%%IMAGE%%" rel="icon"/>
    <style>%%STYLE%%
    </style>
    <title>%%TITLE%%</title>
</head>
<body>
<div class="container">
    <section class="header">
        <img class="value-img" src="%%IMAGE%%" width="150" alt="Repository Image"/>
        <h2 class="title">%%NAME%%<span class="version">%%VERSION%%</span></h2>
    </section>
    %%SUMMARY%%
    <br>
    <pre><code>pip install %%NAME%%%%PIP_VERSION%% --extra-index-url %%URL%%simple</code></pre>
    <hr/>
    <div class="row">
        <section class="side_panel">
            <div class="three columns">
                <h5 class="text-header">Navigation</h5>
                <a class="card" href="#description">Description</a>
                <a class="card" href="#releases">Releases</a>
                <a class="card" href="#files">Files</a>
                <hr/>
                <h5 class="text-header">Project Links</h5>
                <ul>%%LINKS%%
                </ul>
                <hr/>
                <h5 class="text-header">Meta</h5>%%META%%
                <hr/>
                <h5 class="text-header">Classifiers</h5>
                <ul>%%CLASSIFIERS%%
                </ul>
            </div>
        </section>
        <div class="nine columns tabs">
            <div id="releases" class="tab">
                <h6 class="text-header">Releases</h6>%%RELEASES%%

            </div>
            <div id="files" class="tab">
                <h6 class="text-header">Files</h6>%%FILES%%
            </div>
            <div id="description" class="tab">
                <h6 class="text-header">Project Description</h6>
                %%DESCRIPTION%%
            </div>
        </div>
    </div>
    <hr/>
    <section class="footer"></section>
</div>
</body>
</html>
-
